"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const pino = require("express-pino-logger");
const url_path_join_1 = require("../utility/url-path-join");
const core_1 = require("@epandco/unthink-foundation/lib/core");
const express_1 = require("express");
function setHeaders(req, resp, headers) {
    if (!headers) {
        return;
    }
    for (const name in headers) {
        if (name.toLowerCase() === 'content-type') {
            req.log.error('skipping content-type - this cant be set directly');
            continue;
        }
        const currentValue = resp.getHeader(name);
        const newValue = headers[name];
        if (currentValue) {
            req.log.error(`Replacing header value for ${name}. Old: ${currentValue} - New: ${newValue}`);
        }
        resp.set(name, newValue);
    }
}
function setCookie(resp, cookie) {
    const options = {};
    /*
     * Noticed the 'cookie' function below did not like having a full object of undefined properties.
     * The type suggests this is fine but 'maxAge' specifically seems to trip up with the function assuming
     * that if the property exists, even if undefined (per the type) then it should be a number.
     *
     * This feels like a slight rough spot between where types are generated for express but it's obviously not
     * typescript under and they way they check for 'undefined' seems to be if property doesn't exist.
     *
     * The code below creates an empty object and only adds the property to it if the value exist for that property
     * to avoid the problem above.
     */
    if (cookie.domain) {
        options.domain = cookie.domain;
    }
    if (cookie.expires) {
        options.expires = cookie.expires;
    }
    if (cookie.httpOnly) {
        options.httpOnly = cookie.httpOnly;
    }
    if (cookie.maxAge) {
        options.maxAge = cookie.maxAge;
    }
    if (cookie.path) {
        options.path = cookie.path;
    }
    if (cookie.sameSite) {
        options.sameSite = cookie.sameSite;
    }
    if (cookie.secure) {
        options.secure = cookie.secure;
    }
    resp.cookie(cookie.name, cookie.value, options);
}
function setCookies(req, resp, cookies) {
    if (!cookies) {
        return;
    }
    /* If cookies are set then lets not replace existing cookies unless specified by the new cookie config */
    if (req.cookies) {
        for (const cookie of cookies) {
            const currentCookie = req.cookies[cookie.name];
            if (currentCookie && !cookie.overwrite) {
                req.log.error(`Keep existing cookie: ${cookie.name} based on config option to overwrite this cookie.`);
                continue;
            }
            if (currentCookie && cookie.overwrite) {
                req.log.error(`Overwriting cookie: ${cookie.name} based on config option to overwrite this cookie.`);
            }
            setCookie(resp, cookie);
        }
    }
    else {
        /* probably not huge gains but lets not worry about the checks for existing cookies if not exist on the request */
        for (const cookie of cookies) {
            setCookie(resp, cookie);
        }
    }
}
function convertHeaders(req) {
    if (!req.headers) {
        return undefined;
    }
    const headers = {};
    for (const name in req.headers) {
        const value = req.headers[name];
        if (value) {
            // Choosing to ignore the string[] part of this type definition.
            // Not clear why that is needed but can address it when it comes up.
            headers[name] = value;
        }
    }
    return headers;
}
function convertCookies(req) {
    if (!req.cookies) {
        return undefined;
    }
    const cookies = [];
    for (const name in req.cookies) {
        const value = req.cookies[name];
        cookies.push({
            name: name,
            value: value
        });
    }
    return cookies;
}
function buildRouteContext(req, resp) {
    return {
        query: req.query,
        params: req.params,
        body: req.body,
        headers: convertHeaders(req),
        cookies: convertCookies(req),
        logger: req.log,
        local: resp.locals,
        path: req.path
    };
}
function mergeLocals(result, resp) {
    if (result.local) {
        resp.locals = { ...resp.locals, ...result.local };
    }
}
function buildUnthinkError(error) {
    if (error instanceof Error) {
        return { unthinkHandlerError: error.stack };
    }
    return { unthinkHandlerError: JSON.stringify(error) };
}
function redirect(result, req, resp) {
    if ((result.status === 301 || result.status === 302) && result.redirectUrl) {
        setHeaders(req, resp, result.headers);
        setCookies(req, resp, result.cookies);
        resp.redirect(result.status, result.redirectUrl);
        return true;
    }
    if ((result.status === 301 || result.status === 302) && !result.redirectUrl) {
        throw new Error(`When view result has a status of ${result.status} the redirect url MUST BE specified`);
    }
    return false;
}
function buildViewHandler(resourceRouteHandler, render) {
    return async (req, resp, next) => {
        resp.contentType('text/html');
        let error;
        try {
            const ctx = buildRouteContext(req, resp);
            const result = await resourceRouteHandler(ctx);
            // Merging this always because errors may use locals and needs to be passed down
            // to error handler.
            mergeLocals(result, resp);
            if (result.status === 200 && result.template) {
                const body = render(result, 
                // Locals may have changed so rebuild context after merge locals above to ensure
                // render function gets latest locals.
                buildRouteContext(req, resp));
                setHeaders(req, resp, result.headers);
                setCookies(req, resp, result.cookies);
                resp.status(200);
                resp.send(body);
                return;
            }
            if (redirect(result, req, resp)) {
                return;
            }
            if (result.status === 200 && !result.template) {
                error = new Error('When view result has a status of 200 the template MUST BE set!');
            }
            else {
                error = result;
            }
        }
        catch (e) {
            error = e;
        }
        next(error);
    };
}
function buildViewErrorHandler(render) {
    return async (err, req, resp, _next) => {
        if (resp.headersSent) {
            req.log.error('Response already sent. This is likely a bug in the route pipeline in this package.');
            return;
        }
        const unknownErrorMessage = 'Unknown error.';
        if (!err) {
            req.log.error('No error passed into handler');
            resp.status(500).send(unknownErrorMessage);
            return;
        }
        if (!(err instanceof core_1.ViewResult) && !(err instanceof core_1.MiddlewareResult)) {
            req.log.error(buildUnthinkError(err), 'Unexpected error:');
            resp.status(500).send(unknownErrorMessage);
            return;
        }
        const result = err;
        try {
            const view = render(result, buildRouteContext(req, resp));
            setHeaders(req, resp, result.headers);
            setCookies(req, resp, result.cookies);
            resp.status(result.status);
            resp.send(view);
            return;
        }
        catch (err) {
            req.log.error(buildUnthinkError(err), 'Failed to handle result');
            resp.status(500).send(unknownErrorMessage);
        }
    };
}
function buildDataHandler(resourceRouteHandler) {
    return async (req, resp, next) => {
        let error;
        try {
            const ctx = buildRouteContext(req, resp);
            const result = await resourceRouteHandler(ctx);
            // merge locals to use downstream
            mergeLocals(result, resp);
            if (result.status === 200 && result.value) {
                setHeaders(req, resp, result.headers);
                setCookies(req, resp, result.cookies);
                resp.status(result.status).json(result.value);
                return;
            }
            if (result.status === 204 && !result.value) {
                setHeaders(req, resp, result.headers);
                setCookies(req, resp, result.cookies);
                resp.status(204).end();
                return;
            }
            if (redirect(result, req, resp)) {
                return;
            }
            if (result.status === 200 && !result.value) {
                error = new Error('The value MUST be set for data results when the status is 200.');
            }
            else if (result.status === 204 && result.value) {
                error = new Error('The value SHOULD NOT be set for data results when the status is 204');
            }
            else {
                error = result;
            }
        }
        catch (e) {
            error = e;
        }
        next(error);
    };
}
async function dataErrorHandler(err, req, resp, _next) {
    if (resp.headersSent) {
        req.log.error('Response already sent. This is likely a bug in the route pipeline in this package.');
        return;
    }
    const unknownError = 'Unknown error.';
    if (!err) {
        req.log.error('No error passed into dataErrorHandler');
        resp.status(500).json(unknownError);
        return;
    }
    if (!(err instanceof core_1.DataResult) && !(err instanceof core_1.MiddlewareResult)) {
        req.log.error(buildUnthinkError(err), 'Unexpected error');
        resp.status(500).json(unknownError);
        return;
    }
    const result = err;
    if (result.status !== 400 && result.status !== 401 && result.status !== 404) {
        req.log.error(`The status ${result.status} is not supported by this framework for data results.`);
        resp.status(500).json(unknownError);
        return;
    }
    if ((result.status === 401 || result.status === 404) && result.value) {
        req.log.error(`When the data result has a status of ${result.status} the value SHOULD NOT be set.`);
        resp.status(500).json('unknown error');
        return;
    }
    if (result.status === 400 && !result.value) {
        req.log.error('When the data result has a status of 400 the value should MUST be set.');
        resp.status(500).json('unknown error');
        return;
    }
    setHeaders(req, resp, result.headers);
    setCookies(req, resp, result.cookies);
    resp.status(result.status).json(result.value);
}
function dataEndHandler(result, resp) {
    resp.status(result.status).json(result.value);
}
function viewEndHandler(result, resp) {
    resp.contentType('text/html');
    resp.status(result.status).send(result.value);
}
function middlewareHandler(result, endHandler, req, resp, next) {
    mergeLocals(result, resp);
    if (redirect(result, req, resp)) {
        return;
    }
    if (!result.continue && !result.end) {
        next(result);
    }
    setHeaders(req, resp, result.headers);
    setCookies(req, resp, result.cookies);
    if (result.end) {
        return endHandler(result, resp);
    }
    if (result.continue) {
        next();
    }
}
async function asyncMiddlewareHandler(asyncResult, callback) {
    const result = await asyncResult;
    return callback(result);
}
function buildMiddlewareHandler(handler, endHandler) {
    return (req, res, next) => {
        const ctx = buildRouteContext(req, res);
        try {
            const result = handler(ctx);
            if (result instanceof Promise) {
                return asyncMiddlewareHandler(result, (asyncResult) => {
                    middlewareHandler(asyncResult, endHandler, req, res, next);
                });
            }
            middlewareHandler(result, endHandler, req, res, next);
        }
        catch (error) {
            next(error);
        }
    };
}
function wrapUnthinkMiddleware(handler, routeType) {
    if (handler.__middlewareType === core_1.MiddlewareType.RAW) {
        throw new Error('No need to wrap raw middleware');
    }
    switch (routeType) {
        case core_1.RouteType.DATA:
            return buildMiddlewareHandler(handler, dataEndHandler);
            break;
        case core_1.RouteType.VIEW:
            return buildMiddlewareHandler(handler, viewEndHandler);
            break;
        default:
            throw new Error(`Unsupported route type ${routeType}`);
    }
}
function processMiddleware(routeType, middleware) {
    if (!middleware) {
        return undefined;
    }
    let filtered;
    // Only include the relevant middleware for the route type
    // It is not an error for it to be in the list especially from the resource level we just safely ignore it
    if (routeType === core_1.RouteType.DATA) {
        filtered = middleware.filter(p => p.__middlewareType !== core_1.MiddlewareType.VIEW);
    }
    else if (routeType === core_1.RouteType.VIEW) {
        filtered = middleware.filter(p => p.__middlewareType !== core_1.MiddlewareType.DATA);
    }
    else {
        throw new Error(`Route type ${routeType} is not supported`);
    }
    return filtered.map((handler) => {
        if (handler.__middlewareType !== core_1.MiddlewareType.RAW) {
            return wrapUnthinkMiddleware(handler, routeType);
        }
        if ('__expressMiddleware' in handler && handler.__expressMiddleware === 'EXPRESS_MIDDLEWARE') {
            return handler;
        }
        throw new Error(`Unsupported middleware ${JSON.stringify(handler)}`);
    });
}
class UnthinkExpressGenerator {
    constructor(app, viewRenderer, logLevel) {
        this.app = app;
        this.viewRenderer = viewRenderer;
        this.logLevel = logLevel;
    }
    generate(resourceDefinitions) {
        const generatedDefinitions = resourceDefinitions.flatMap(p => this.generateDefinition(p));
        this.app.use(express_1.json());
        this.app.use(pino({
            level: this.logLevel
        }));
        for (const { path, router } of generatedDefinitions) {
            this.app.use(path, router);
        }
    }
    generateRoute(resourceRouteDefinition, resourceMiddleware) {
        var _a;
        const router = express_1.Router();
        const route = router.route(resourceRouteDefinition.path);
        switch (resourceRouteDefinition.__routeType) {
            case 'DATA':
                router.use(dataErrorHandler);
                break;
            case 'VIEW':
                router.use(buildViewErrorHandler(this.viewRenderer));
                break;
        }
        const routeAndResourceLevelHandlers = [];
        // Inject route type into the middleware pipeline in case raw middleware needs it.
        routeAndResourceLevelHandlers.push((_req, resp, next) => {
            resp.locals.__routeType = resourceRouteDefinition.__routeType;
            next();
        });
        const processedResourceMiddleware = processMiddleware(resourceRouteDefinition.__routeType, resourceMiddleware);
        if (processedResourceMiddleware) {
            routeAndResourceLevelHandlers.push(...processedResourceMiddleware);
        }
        const processedRouteMiddleware = processMiddleware(resourceRouteDefinition.__routeType, resourceRouteDefinition.middleware);
        if (processedRouteMiddleware) {
            routeAndResourceLevelHandlers.push(...processedRouteMiddleware);
        }
        route.all(routeAndResourceLevelHandlers);
        for (const method in resourceRouteDefinition.methods) {
            const resourceHandlerObj = resourceRouteDefinition.methods[method];
            if (!resourceHandlerObj) {
                throw new Error('Handler must be defined.');
            }
            let resourceHandler;
            const handlers = [];
            if ('handler' in resourceHandlerObj) {
                resourceHandler = resourceHandlerObj.handler;
                const processedHandlerMiddleware = processMiddleware(resourceRouteDefinition.__routeType, resourceHandlerObj.middleware);
                if (processedHandlerMiddleware) {
                    handlers.push(...processedHandlerMiddleware);
                }
            }
            else {
                resourceHandler = resourceHandlerObj;
            }
            switch (resourceRouteDefinition.__routeType) {
                case 'DATA':
                    handlers.push(buildDataHandler(resourceHandler));
                    break;
                case 'VIEW':
                    handlers.push(buildViewHandler(resourceHandler, this.viewRenderer));
                    break;
            }
            route[method](...handlers);
        }
        return {
            prefix: (_a = resourceRouteDefinition.prefix) !== null && _a !== void 0 ? _a : '',
            router: router
        };
    }
    generateDefinition(resourceDefinition) {
        var _a;
        const mapPrefixedRoutes = new Map();
        for (const routeDef of resourceDefinition.routes) {
            const { prefix, router } = this.generateRoute(routeDef, resourceDefinition.middleware);
            if (!mapPrefixedRoutes.has(prefix)) {
                let basePath = resourceDefinition.basePath;
                if (!basePath) {
                    console.log(`basePath is not defined for ${resourceDefinition.name} and being defaulted to '/'`);
                    basePath = '/';
                }
                const urlPath = url_path_join_1.urlPathJoin([prefix ? prefix : '/', basePath]);
                const generatedDefinition = {
                    path: urlPath,
                    router: express_1.Router()
                };
                mapPrefixedRoutes.set(prefix, generatedDefinition);
            }
            (_a = mapPrefixedRoutes.get(prefix)) === null || _a === void 0 ? void 0 : _a.router.use(router);
        }
        return Array.from(mapPrefixedRoutes.values());
    }
}
exports.UnthinkExpressGenerator = UnthinkExpressGenerator;
//# sourceMappingURL=unthink-express-generator.js.map